/* $Id: libperfctr.c,v 1.7 2000/08/31 22:32:06 mikpe Exp $
 * Library interface to Linux x86 Performance-Monitoring Counters.
 *
 * Copyright (C) 1999-2000  Mikael Pettersson
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "libperfctr.h"

#define PAGE_SIZE	4096

static unsigned char cpu_type;
static int have_rdpmc = 0;	/* false on pre-MMX Intel P5 */
static int have_rdtsc = 0;	/* false on WinChip when PMCs are used */
static int dev_perfctr_fd = -1;
static unsigned long cpu_khz;

struct vperfctr {
    int fd;
    volatile const struct vperfctr_state *kstate;
};

static int check_init_unlocked(void)
{
    if( dev_perfctr_fd < 0 ) {
	struct perfctr_info info;

	if( (dev_perfctr_fd = open("/dev/perfctr", O_RDONLY)) < 0 ) {
	    perror("open /dev/perfctr");
	    return -1;
	}
	if( ioctl(dev_perfctr_fd, PERFCTR_INFO, &info) < 0 )
	    return -1;
	cpu_type = info.cpu_type;
	have_rdpmc = info.cpu_features & PERFCTR_FEATURE_RDPMC;
	have_rdtsc = info.cpu_features & PERFCTR_FEATURE_RDTSC;
	cpu_khz = info.cpu_khz;
    }
    return 1;
}

#if defined(USE_PTHREAD)
#include <pthread.h>
static int check_init_locked(void)
{
    static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
    int result;

    pthread_mutex_lock(&lock);
    result = check_init_unlocked();
    pthread_mutex_unlock(&lock);
    return result;
}
#define check_init()	check_init_locked()
#else
#define check_init()	check_init_unlocked()
#endif

/*
 * Invokes the kernel driver directly.
 */

int perfctr_syscall(unsigned cmd, long arg)
{
    if( check_init() < 0 )
	return -1;
    return ioctl(dev_perfctr_fd, cmd, arg);
}

/*
 * Operations on the process' own perfctrs.
 */

struct vperfctr *vperfctr_attach(void)
{
    int oerrno;
    struct vperfctr *perfctr;

    perfctr = malloc(sizeof(struct vperfctr));
    if( !perfctr )
	return perfctr;
    perfctr->fd = perfctr_syscall(VPERFCTR_ATTACH, 0);
    if( perfctr->fd >= 0 ) {
	 perfctr->kstate = mmap(NULL, PAGE_SIZE, PROT_READ,
				MAP_SHARED, perfctr->fd, 0);
	 if( perfctr->kstate != MAP_FAILED )
	      return perfctr;
	 oerrno = errno;
	 close(perfctr->fd);
	 errno = oerrno;
    }
    oerrno = errno;
    free(perfctr);
    errno = oerrno;
    return NULL;
}

#define rdtscl(low)	\
	__asm__ __volatile__("rdtsc" : "=a"(low) : : "edx")
#define rdpmcl(ctr,low)	\
	__asm__ __volatile__("rdpmc" : "=a"(low) : "c"(ctr) : "edx")

unsigned long long vperfctr_read_one(const struct vperfctr *self, int i)
{
    unsigned long long sum;
    unsigned int start, now;
    unsigned int tsc0, tsc1;

    if( i > 0 && !have_rdpmc ) {	/* Intel pre-MMX P5 lossage */
	ioctl(self->fd, VPERFCTR_SAMPLE, NULL);
	return self->kstate->sum.ctr[i];
    }
    tsc1 = self->kstate->start.ctr[0];
    if( i == 0 ) {	/* Retrieve the virtualised TSC. */
	do {
	    tsc0 = tsc1;
	    rdtscl(now);
	    sum = self->kstate->sum.ctr[0];
	    tsc1 = self->kstate->start.ctr[0];
	} while( tsc0 != tsc1 );
	start = tsc1;
    } else {		/* Retrieve the virtualised PMC #(i-1). */
	do {
	    tsc0 = tsc1;
	    rdpmcl(i-1, now);
	    sum = self->kstate->sum.ctr[i];
	    start = self->kstate->start.ctr[i];
	    tsc1 = self->kstate->start.ctr[0];
	} while( tsc0 != tsc1 );
    }
    return sum + (now - start);
}

int vperfctr_read_state(const struct vperfctr *self, struct vperfctr_state *state)
{
    unsigned int prev_tsc, next_tsc;
    struct perfctr_low_ctrs now;
    int i, nrctrs;

    /* If the counters are stopped, or if at least one PMC is enabled
     * but the CPU doesn't support RDPMC, then tell the kernel to sync
     * its state, and copy it from the mmap() buffer.
     */
    nrctrs = self->kstate->status;
    if( (nrctrs > 1 && !have_rdpmc) || nrctrs <= 0 ) {
	ioctl(self->fd, VPERFCTR_SAMPLE, NULL);
	do {
	    *state = *self->kstate;
	} while( state->start.ctr[0] != self->kstate->start.ctr[0] );
	return 0;
    }
    /* initial context-switch timestamp */
    next_tsc = self->kstate->start.ctr[0];
    do {
	prev_tsc = next_tsc;
	/* read register contents */
	if( have_rdtsc )	/* WinChip lossage */
	    rdtscl(now.ctr[0]);
	for(i = nrctrs - 1; --i >= 0;)
	    rdpmcl(i, now.ctr[i+1]); /* TSC is ctr[0], PMC i is ctr[i+1] */
	/* copy kernel state */
	*state = *self->kstate;
	/* next context-switch timestamp */
	next_tsc = self->kstate->start.ctr[0];
	/* loop until the copy is consistent (the timestamps match) */
    } while( prev_tsc != next_tsc );
    /* now we have a good copy; update the sums */
    if( !have_rdtsc )		/* WinChip lossage */
	now.ctr[0] = 0;
    for(i = 0; i < nrctrs; ++i)
	state->sum.ctr[i] += now.ctr[i] - state->start.ctr[i];
    /* done */
    return 0;
}

int vperfctr_control(const struct vperfctr *perfctr,
		     struct perfctr_control *control)
{
    return ioctl(perfctr->fd, VPERFCTR_CONTROL, control);
}

int vperfctr_stop(const struct vperfctr *perfctr)
{
    return ioctl(perfctr->fd, VPERFCTR_STOP, NULL);
}

int vperfctr_unlink(const struct vperfctr *perfctr)
{
    return ioctl(perfctr->fd, VPERFCTR_UNLINK, NULL);
}

void vperfctr_close(struct vperfctr *perfctr)
{
    if( perfctr->fd >= 0 ) {
	munmap((void*)perfctr->kstate, PAGE_SIZE);
	close(perfctr->fd);
	perfctr->kstate = NULL;
	perfctr->fd = -1;
    }
    free(perfctr);
}

/*
 * Operations on global-mode perfctrs.
 */

int perfctr_global_control(struct gperfctr_control *arg)
{
    return perfctr_syscall(GPERFCTR_CONTROL, (long)arg);
}

int perfctr_global_read(struct gperfctr_state *arg)
{
    return perfctr_syscall(GPERFCTR_READ, (long)arg);
}

int perfctr_global_stop(void)
{
    return perfctr_syscall(GPERFCTR_STOP, 0);
}

/*
 * Miscellaneous operations.
 */

int perfctr_info(struct perfctr_info *info)
{
    return perfctr_syscall(PERFCTR_INFO, (long)info);
}

unsigned perfctr_cpu_nrctrs(void)
{
    if( check_init() < 0 )
	return 0;
    switch( cpu_type ) {
      case PERFCTR_X86_GENERIC:
	return 1;
      case PERFCTR_X86_AMD_K7:
	return 5;
      default:
	return 3;
    }
}

const char *perfctr_cpu_name(void)
{
    if( check_init() < 0 )
	return NULL;
    switch( cpu_type ) {
      case PERFCTR_X86_GENERIC:
	return "Generic x86 with TSC";
      case PERFCTR_X86_INTEL_P5:
        return "Intel Pentium";
      case PERFCTR_X86_INTEL_P5MMX:
        return "Intel Pentium MMX";
      case PERFCTR_X86_INTEL_P6:
        return "Intel Pentium Pro";
      case PERFCTR_X86_INTEL_PII:
        return "Intel Pentium II";
      case PERFCTR_X86_INTEL_PIII:
        return "Intel Pentium III";
      case PERFCTR_X86_CYRIX_MII:
        return "Cyrix 6x86MX/MII/III";
      case PERFCTR_X86_WINCHIP_C6:
	return "WinChip C6";
      case PERFCTR_X86_WINCHIP_2:
	return "WinChip 2";
      case PERFCTR_X86_AMD_K7:
	return "AMD K7 Athlon";
      default:
        return "?";
    }
}

unsigned perfctr_evntsel_num_insns(void)
{
    /* This is terribly naive. Assumes only one event will be
     * selected, at CPL > 0.
     */
    if( check_init() < 0 )
	return (unsigned)-1;
    switch( cpu_type ) {
      case PERFCTR_X86_INTEL_P5:
      case PERFCTR_X86_INTEL_P5MMX:
      case PERFCTR_X86_CYRIX_MII:
        /* event 0x16, count at CPL 3 */
        return 0x16 | (2 << 6);
      case PERFCTR_X86_INTEL_P6:
      case PERFCTR_X86_INTEL_PII:
      case PERFCTR_X86_INTEL_PIII:
      case PERFCTR_X86_AMD_K7:
        /* event 0xC0, count at CPL > 0, ENable */
        return 0xC0 | (1 << 16) | (1 << 22);
      case PERFCTR_X86_WINCHIP_C6:
        return 0x02;
      case PERFCTR_X86_WINCHIP_2:
        return 0x16;
      default:
        fprintf(stderr, __FUNCTION__ ": cpu_type %u not recognized\n", cpu_type);
        exit(1);
    }
}
